<a class="button button-with-icon" href="https://github.com/bazelbuild/bazel/issues/new?title=%5Bbazel.build%5D+Problem+with+{% dynamic print request.path %}&template=doc_issue.yml&link=https%3A%2F%2Fbazel.build{% dynamic print request.path %}" target="_blank">
  Report an issue<span class="material-icons icon-after" aria-hidden="true">open_in_new</span>
</a>
{% dynamic if not setvar.source_file %}
<a class="button button-with-icon" href="https://github.com/bazelbuild/bazel/tree/master/site/en{% dynamic print request.path %}.md" target="_blank">
  View source<span class="material-icons icon-after" aria-hidden="true">open_in_new</span>
</a>
{% dynamic elif setvar.source_file != "NONE" %}
<a class="button button-with-icon" href="https://github.com/bazelbuild/bazel/tree/master/{% dynamic print setvar.source_file %}" target="_blank">
  View source<span class="material-icons icon-after" aria-hidden="true">open_in_new</span>
</a>
{% dynamic endif %}
{% dynamic if not setvar.original_path %}
{% dynamic setvar original_path %}{% dynamic print request.path %}{% dynamic endsetvar %}
{% dynamic endif %}
<span style="float: right; line-height: 36px">
{% dynamic if not setvar.version %}
<strong>Nightly</strong>
{% dynamic else %}
<a href="{% dynamic print setvar.original_path %}">Nightly</a>
{% dynamic endif %}
<!-- The lines below are updated by //scripts/docs:gen_new_toc -->
<!-- BEGIN_VERSION_INDICATOR -->
·
{% dynamic if setvar.version == "7.2.0" %}
<strong>7.2</strong>
{% dynamic else %}
<a href="/versions/7.2.0{% dynamic print setvar.original_path %}">7.2</a>
{% dynamic endif %}
·
{% dynamic if setvar.version == "7.1.0" %}
<strong>7.1</strong>
{% dynamic else %}
<a href="/versions/7.1.0{% dynamic print setvar.original_path %}">7.1</a>
{% dynamic endif %}
·
{% dynamic if setvar.version == "7.0.0" %}
<strong>7.0</strong>
{% dynamic else %}
<a href="/versions/7.0.0{% dynamic print setvar.original_path %}">7.0</a>
{% dynamic endif %}
·
{% dynamic if setvar.version == "6.5.0" %}
<strong>6.5</strong>
{% dynamic else %}
<a href="/versions/6.5.0{% dynamic print setvar.original_path %}">6.5</a>
{% dynamic endif %}
·
{% dynamic if setvar.version == "6.4.0" %}
<strong>6.4</strong>
{% dynamic else %}
<a href="/versions/6.4.0{% dynamic print setvar.original_path %}">6.4</a>
{% dynamic endif %}
<!-- END_VERSION_INDICATOR -->
</span>
</p>
