<p>List of <a href="${link build-ref#labels}">labels</a>;
  <a href="#configurable-attributes">nonconfigurable</a>;
  default is <code><a href="${link package.default_visibility}">default_visibility</a></code> from
  <a href="${link package}">package</a> if specified, or <code>"//visibility:private"</code>
  otherwise</p>

<p>
The <code>visibility</code> attribute on a target controls whether the target
can be used in other packages. See the documentation for
<a href="${link visibility}">visibility</a>.
</p>
