// Copyright 2016 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

syntax = "proto3";

package blaze;

// option java_api_version = 2;
option java_package = "com.google.devtools.build.lib.rules.android.deployinfo";

message Artifact {
  // This path is relative to "bazel info execution_root"
  string exec_root_path = 1;

  // Points to a globally valid URI whenever possible.
  // See BuildEventStreamOptions.java
  string canonical_uri = 2;
}

// Information about how to deploy and launch the result of an android build
message AndroidDeployInfo {
  // The merged manifest in the resource APK
  Artifact merged_manifest = 1;

  // Merged manifests of dependencies
  repeated Artifact additional_merged_manifests = 2;

  // If the build supports an external deploy, this will have the APKs.
  // mobile-install does its own deployment, so this list will be empty
  repeated Artifact apks_to_deploy = 3;

  // Unused
  repeated Artifact data_to_deploy = 4;

  // Used by mobile-install
  string manifest_package_name = 5;
}
