// Copyright 2017 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.google.devtools.build.lib.rules.android;

import com.google.common.collect.ImmutableList;
import com.google.devtools.build.lib.actions.Artifact;
import com.google.devtools.build.lib.collect.nestedset.NestedSet;
import com.google.devtools.build.lib.concurrent.ThreadSafety.Immutable;
import com.google.devtools.build.lib.packages.BuiltinProvider;
import com.google.devtools.build.lib.packages.NativeInfo;

/**
 * Information about an {@code android_host_service_fixture} to run as part of an {@code
 * android_instrumentation_test}.
 */
@Immutable
public class AndroidHostServiceFixtureInfoProvider extends NativeInfo {

  private static final String STARLARK_NAME = "HostServiceFixtureInfo";
  static final BuiltinProvider<AndroidHostServiceFixtureInfoProvider>
      ANDROID_HOST_SERVICE_FIXTURE_INFO =
          new BuiltinProvider<AndroidHostServiceFixtureInfoProvider>(
              STARLARK_NAME, AndroidHostServiceFixtureInfoProvider.class) {};

  private final Artifact executable;
  private final ImmutableList<String> serviceNames;
  private final NestedSet<Artifact> supportApks;
  private final boolean providesTestArgs;
  private final boolean daemon;

  AndroidHostServiceFixtureInfoProvider(
      Artifact executable,
      ImmutableList<String> serviceNames,
      NestedSet<Artifact> supportApks,
      boolean providesTestArgs,
      boolean isDaemon) {
    this.executable = executable;
    this.serviceNames = serviceNames;
    this.supportApks = supportApks;
    this.providesTestArgs = providesTestArgs;
    this.daemon = isDaemon;
  }

  @Override
  public BuiltinProvider<AndroidHostServiceFixtureInfoProvider> getProvider() {
    return ANDROID_HOST_SERVICE_FIXTURE_INFO;
  }

  public Artifact getExecutable() {
    return executable;
  }

  public ImmutableList<String> getServiceNames() {
    return serviceNames;
  }

  public NestedSet<Artifact> getSupportApks() {
    return supportApks;
  }

  public boolean getProvidesTestArgs() {
    return providesTestArgs;
  }

  public boolean getDaemon() {
    return daemon;
  }
}
