// Copyright 2020 The Bazel Authors. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//    http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package com.google.devtools.build.lib.analysis;

import com.google.devtools.build.lib.analysis.config.BuildOptions;
import com.google.devtools.common.options.OptionDefinition;

/**
 * Predicate for determining whether the analysis cache should be cleared, given the new and old
 * value of an option which has changed and the values of the other new options.
 */
@FunctionalInterface
public interface OptionsDiffPredicate {
  OptionsDiffPredicate ALWAYS_INVALIDATE = (options, option, oldValue, newValue) -> true;

  boolean apply(BuildOptions newOptions, OptionDefinition option, Object oldValue, Object newValue);
}
