/*
 * Copyright 2020 The Bazel Authors. All rights reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.devtools.build.android.desugar.io;

import com.google.common.io.Resources;
import com.google.devtools.build.android.desugar.langmodel.ClassName;
import java.io.IOError;
import java.io.IOException;
import java.io.InputStream;

/** Provides class file content based on resource files. */
public class ResourceBasedClassFiles implements ClassFileBatchProvider {

  @Override
  public FileContentProvider<InputStream> getContent(ClassName className) {
    String inArchiveBinaryPathName = className.classFilePathName();
    return new FileContentProvider<>(
        inArchiveBinaryPathName,
        () -> {
          try {
            return Resources.getResource(inArchiveBinaryPathName).openStream();
          } catch (IOException e) {
            throw new IOError(e);
          }
        });
  }
}
